bugs_by_version = {
    "0.1.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.7": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.3.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
    ],
    "0.3.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
    ],
    "0.3.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
    ],
    "0.3.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
    ],
    "0.4.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "LibrariesNotCallableFromPayableFunctions",
    ],
    "0.4.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "LibrariesNotCallableFromPayableFunctions",
    ],
    "0.4.10": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.11": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
    ],
    "0.4.12": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
    ],
    "0.4.13": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
    ],
    "0.4.14": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
    ],
    "0.4.15": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.16": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.17": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.18": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.19": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
    ],
    "0.4.20": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.21": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.22": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "OneOfTwoConstructorsSkipped",
    ],
    "0.4.23": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
    ],
    "0.4.24": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
    ],
    "0.4.25": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
    ],
    "0.4.26": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
    ],
    "0.4.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
    ],
    "0.4.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
    ],
    "0.4.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStateKnowledgeNotResetForJumpdest",
    ],
    "0.4.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
    ],
    "0.4.7": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.8": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.9": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.5.0": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.1": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.10": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
    ],
    "0.5.11": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.12": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.13": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.14": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2LoopYulOptimizer",
    ],
    "0.5.15": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.16": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
    ],
    "0.5.17": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.5.2": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.3": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.4": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.5": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
        "IncorrectByteInstructionOptimization",
        "DoubleShiftSizeOverflow",
    ],
    "0.5.6": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
        "IncorrectByteInstructionOptimization",
    ],
    "0.5.7": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
    ],
    "0.5.8": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
    ],
    "0.5.9": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
    ],
    "0.6.0": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "YulOptimizerRedundantAssignmentBreakContinue",
    ],
    "0.6.1": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.10": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.11": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.12": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.2": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.3": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.4": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.5": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
    ],
    "0.6.6": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
    ],
    "0.6.7": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
    ],
    "0.6.8": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.9": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "UsingForCalldata",
    ],
    "0.7.0": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.7.1": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "FreeFunctionRedefinition",
    ],
    "0.7.2": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.7.3": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
    ],
    "0.7.4": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.7.5": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.7.6": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.0": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.1": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.10": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
    "0.8.11": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "AbiEncodeCallLiteralAsFixedBytesBug",
    ],
    "0.8.12": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "AbiEncodeCallLiteralAsFixedBytesBug",
    ],
    "0.8.13": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "InlineAssemblyMemorySideEffects",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
    "0.8.14": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "InlineAssemblyMemorySideEffects",
    ],
    "0.8.15": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
    ],
    "0.8.16": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
    ],
    "0.8.17": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.18": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.19": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.2": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.20": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.21": ["VerbatimInvalidDeduplication"],
    "0.8.22": ["VerbatimInvalidDeduplication"],
    "0.8.23": [],
    "0.8.24": [],
    "0.8.25": [],
    "0.8.3": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
    ],
    "0.8.4": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.5": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.6": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.7": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.8": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "UserDefinedValueTypesBug",
        "SignedImmutables",
    ],
    "0.8.9": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
}



"""
Module printing summary of the contract
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.core.declarations.function import Function
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.output import Output


class PrinterWrittenVariablesAndAuthorization(AbstractPrinter):
    ARGUMENT = "vars-and-auth"
    HELP = "Print the state variables written and the authorization of the functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#variables-written-and-authorization"

    @staticmethod
    def get_msg_sender_checks(function: Function) -> list[str]:
        all_functions = [
            ir.function for ir in function.all_internal_calls() if isinstance(ir.function, Function)
        ] + [function]

        all_nodes_ = [f.nodes for f in all_functions]
        all_nodes = [item for sublist in all_nodes_ for item in sublist]

        all_conditional_nodes = [
            n for n in all_nodes if n.contains_if() or n.contains_require_or_assert()
        ]
        all_conditional_nodes_on_msg_sender = [
            str(n.expression)
            for n in all_conditional_nodes
            if "msg.sender" in [v.name for v in n.solidity_variables_read]
        ]
        return all_conditional_nodes_on_msg_sender

    def output(self, _filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        all_tables = []
        for contract in self.contracts:  # type: ignore
            txt += f"\nContract {contract.name}\n"
            table = MyPrettyTable(
                ["Function", "State variables written", "Conditions on msg.sender"]
            )
            for function in contract.functions:
                state_variables_written = [
                    v.name for v in function.all_state_variables_written() if v.name
                ]
                msg_sender_condition = self.get_msg_sender_checks(function)
                table.add_row(
                    [
                        function.name,
                        str(sorted(state_variables_written)),
                        str(sorted(msg_sender_condition)),
                    ]
                )
            all_tables.append((contract.name, table))
            txt += str(table) + "\n"

        self.info(txt)
        res = self.generate_output(txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from typing import TYPE_CHECKING

from slither.core.declarations.top_level import TopLevel
from slither.utils.using_for import USING_FOR

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope


class UsingForTopLevel(TopLevel):
    def __init__(self, scope: "FileScope") -> None:
        super().__init__()
        self._using_for: USING_FOR = {}
        self.file_scope: FileScope = scope

    @property
    def using_for(self) -> USING_FOR:
        return self._using_for



"""
Module detecting missing zero address validation

"""

from collections import defaultdict

from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function import ModifierStatements
from slither.core.declarations.function_contract import FunctionContract
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.variables.local_variable import LocalVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Call
from slither.slithir.operations import Send, Transfer, LowLevelCall
from slither.utils.output import Output


class MissingZeroAddressValidation(AbstractDetector):
    """
    Missing zero address validation
    """

    ARGUMENT = "missing-zero-check"
    HELP = "Missing Zero Address Validation"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#missing-zero-address-validation"
    WIKI_TITLE = "Missing zero address validation"
    WIKI_DESCRIPTION = "Detect missing zero address validation."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {

  modifier onlyAdmin {
    if (msg.sender != owner) throw;
    _;
  }

  function updateOwner(address newOwner) onlyAdmin external {
    owner = newOwner;
  }
}
```
Bob calls `updateOwner` without specifying the `newOwner`, so Bob loses ownership of the contract.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Check that the address is not zero."

    def _zero_address_validation_in_modifier(
        self, var: LocalVariable, modifier_exprs: list[ModifierStatements]
    ) -> bool:
        for mod in modifier_exprs:
            for node in mod.nodes:
                # Skip validation if the modifier's parameters contains more than one variable
                # For example
                # function f(a) my_modif(some_internal_function(a, b)) {
                if len(node.irs) != 1:
                    continue
                args = [arg for ir in node.irs if isinstance(ir, Call) for arg in ir.arguments]
                # Check in modifier call arguments and then identify validation of corresponding parameter within modifier context
                if var in args and self._zero_address_validation(
                    mod.modifier.parameters[args.index(var)], mod.modifier.nodes[-1], []
                ):
                    return True
        return False

    def _zero_address_validation(
        self, var: LocalVariable, node: Node, explored: list[Node]
    ) -> bool:
        """
        Detects (recursively) if var is (zero address) checked in the function node
        """
        if node in explored:
            return False
        explored.append(node)

        # Heuristic: Assume zero address checked if variable is used within conditional or require/assert
        # TBD: Actually check for zero address in predicate
        if (node.contains_if() or node.contains_require_or_assert()) and (
            var in node.variables_read
        ):
            return True

        # Check recursively in all the parent nodes
        for father in node.fathers:
            if self._zero_address_validation(var, father, explored):
                return True
        return False

    def _detect_missing_zero_address_validation(
        self, contract: Contract
    ) -> list[tuple[FunctionContract, defaultdict[LocalVariable, list[Node]]]]:
        """
        Detects if addresses are zero address validated before use.
        :param contract: The contract to check
        :return: Functions with nodes where addresses used are not zero address validated earlier
        """
        results = []

        for function in contract.functions_entry_points:
            var_nodes = defaultdict(list)

            for node in function.nodes:
                sv_addrs_written = [
                    sv
                    for sv in node.state_variables_written
                    if sv.type == ElementaryType("address")
                ]

                addr_calls = False
                for ir in node.irs:
                    if isinstance(ir, (Send, Transfer, LowLevelCall)):
                        addr_calls = True

                # Continue if no address-typed state variables are written and if no send/transfer/call
                if not sv_addrs_written and not addr_calls:
                    continue

                # Check local variables used in such nodes
                for var in node.local_variables_read:
                    # Check for address types that are tainted but not by msg.sender
                    if var.type == ElementaryType("address") and is_tainted(
                        var, function, ignore_generic_taint=True
                    ):
                        # Check for zero address validation of variable
                        # in the context of modifiers used or prior function context
                        if not (
                            self._zero_address_validation_in_modifier(
                                var, function.modifiers_statements
                            )
                            or self._zero_address_validation(var, node, [])
                        ):
                            # Report a variable only once per function
                            var_nodes[var].append(node)
            if var_nodes:
                results.append((function, var_nodes))
        return results

    def _detect(self) -> list[Output]:
        """Detect if addresses are zero address validated before use.
        Returns:
            list: {'(function, node)'}
        """

        # Check derived contracts for missing zero address validation
        results = []
        for contract in self.compilation_unit.contracts_derived:
            missing_zero_address_validation = self._detect_missing_zero_address_validation(contract)
            for _, var_nodes in missing_zero_address_validation:
                for var, nodes in var_nodes.items():
                    info: DETECTOR_INFO = [var, " lacks a zero-check on ", ":\n"]
                    for node in nodes:
                        info += ["\t\t- ", node, "\n"]
                    res = self.generate_result(info)
                    results.append(res)
        return results



from slither.tools.properties.properties.properties import (
    Property,
    PropertyType,
    PropertyReturn,
    PropertyCaller,
)

ERC20_Transferable = [
    Property(
        name="crytic_zero_always_empty_ERC20Properties()",
        description="The address 0x0 should not receive tokens.",
        content="""
\t\treturn this.balanceOf(address(0x0)) == 0;""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="crytic_approve_overwrites()",
        description="Allowance can be changed.",
        content="""
\t\tbool approve_return;
\t\tapprove_return = approve(crytic_user, 10);
\t\trequire(approve_return);
\t\tapprove_return = approve(crytic_user, 20);
\t\trequire(approve_return);
\t\treturn this.allowance(msg.sender, crytic_user) == 20;""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_less_than_total_ERC20Properties()",
        description="Balance of one user must be less or equal to the total supply.",
        content="""
\t\treturn this.balanceOf(msg.sender) <= this.totalSupply();""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_totalSupply_consistant_ERC20Properties()",
        description="Balance of the crytic users must be less or equal to the total supply.",
        content="""
\t\treturn this.balanceOf(crytic_owner) + this.balanceOf(crytic_user) + this.balanceOf(crytic_attacker) <= this.totalSupply();""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="crytic_revert_transfer_to_zero_ERC20PropertiesTransferable()",
        description="Using transfer to send tokens to the address 0x0 will revert.",
        content="""
\t\tif (this.balanceOf(msg.sender) == 0){
\t\t\trevert();
\t\t}
\t\treturn transfer(address(0x0), this.balanceOf(msg.sender));""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.FAIL_OR_THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_revert_transferFrom_to_zero_ERC20PropertiesTransferable()",
        description="Using transferFrom to send tokens to the address 0x0 will revert.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\tif (balance == 0){
\t\t\trevert();
\t\t}
\t\tapprove(msg.sender, balance);
\t\treturn transferFrom(msg.sender, address(0x0), this.balanceOf(msg.sender));""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.FAIL_OR_THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_self_transferFrom_ERC20PropertiesTransferable()",
        description="Self transfering tokens using transferFrom works as expected.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\tbool approve_return = approve(msg.sender, balance);
\t\tbool transfer_return = transferFrom(msg.sender, msg.sender, balance);
\t\treturn (this.balanceOf(msg.sender) == balance) && approve_return && transfer_return;""",
        type=PropertyType.HIGH_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_self_transferFrom_to_other_ERC20PropertiesTransferable()",
        description="Transfering tokens to other address using transferFrom works as expected.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\tbool approve_return = approve(msg.sender, balance);
\t\taddress other = crytic_user;
\t\tif (other == msg.sender) {
\t\t\tother = crytic_owner;
\t\t}
\t\tbool transfer_return = transferFrom(msg.sender, other, balance);
\t\treturn (this.balanceOf(msg.sender) == 0) && approve_return && transfer_return;""",
        type=PropertyType.HIGH_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_self_transfer_ERC20PropertiesTransferable()",
        description="Self transfering tokens using transfer works as expected.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\tbool transfer_return = transfer(msg.sender, balance);
\t\treturn (this.balanceOf(msg.sender) == balance) && transfer_return;""",
        type=PropertyType.HIGH_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_transfer_to_other_ERC20PropertiesTransferable()",
        description="Transfering tokens to other address using transfer works as expected.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\taddress other = crytic_user;
\t\tif (other == msg.sender) {
\t\t\tother = crytic_owner;
\t\t}
\t\tif (balance >= 1) {
\t\t\tbool transfer_other = transfer(other, 1);
\t\t\treturn (this.balanceOf(msg.sender) == balance-1) && (this.balanceOf(other) >= 1) && transfer_other;
\t\t}
\t\treturn true;""",
        type=PropertyType.HIGH_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_revert_transfer_to_user_ERC20PropertiesTransferable()",
        description="Transfering more tokens than the balance will revert.",
        content="""
\t\tuint balance = this.balanceOf(msg.sender);
\t\tif (balance == (2 ** 256 - 1))
\t\t\treturn true;
\t\tbool transfer_other = transfer(crytic_user, balance+1);
\t\treturn transfer_other;""",
        type=PropertyType.HIGH_SEVERITY,
        return_type=PropertyReturn.FAIL_OR_THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
]


ERC20_Pausable = [
    Property(
        name="crytic_revert_transfer_ERC20AlwaysTruePropertiesNotTransferable()",
        description="Cannot transfer.",
        content="""
\t\treturn transfer(crytic_user, this.balanceOf(msg.sender));""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.FAIL_OR_THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_revert_transferFrom_ERC20AlwaysTruePropertiesNotTransferable()",
        description="Cannot execute transferFrom.",
        content="""
\t\tapprove(msg.sender, this.balanceOf(msg.sender));
\t\ttransferFrom(msg.sender, msg.sender, this.balanceOf(msg.sender));""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.FAIL_OR_THROW,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_constantBalance()",
        description="Cannot change the balance.",
        content="""
\t\treturn this.balanceOf(crytic_user) == initialBalance_user && this.balanceOf(crytic_attacker) == initialBalance_attacker;""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
    Property(
        name="crytic_constantAllowance()",
        description="Cannot change the allowance.",
        content="""
\t\treturn (this.allowance(crytic_user, crytic_attacker) == initialAllowance_user_attacker) &&
\t\t\t(this.allowance(crytic_attacker, crytic_attacker) == initialAllowance_attacker_attacker);""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ALL,
    ),
]



import abc
import json
from typing import Optional

from slither.core.cfg.node import NodeType, Node, link_nodes
from slither.core.cfg.scope import Scope
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import (
    Function,
    SolidityFunction,
    Contract,
)
from slither.core.declarations.function import FunctionLanguage
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.expressions import (
    Literal,
    AssignmentOperation,
    AssignmentOperationType,
    Identifier,
    CallExpression,
    TupleExpression,
    BinaryOperation,
    UnaryOperation,
)
from slither.core.expressions.expression import Expression
from slither.core.scope.scope import FileScope
from slither.core.solidity_types import ElementaryType
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.local_variable import LocalVariable
from slither.exceptions import SlitherException
from slither.solc_parsing.yul.evm_functions import (
    format_function_descriptor,
    builtins,
    YulBuiltin,
    unary_ops,
    binary_ops,
)
from slither.solc_parsing.expressions.find_variable import find_top_level
from slither.visitors.expression.find_calls import FindCalls
from slither.visitors.expression.read_var import ReadVar
from slither.visitors.expression.write_var import WriteVar


class YulNode:
    def __init__(self, node: Node, scope: "YulScope") -> None:
        self._node = node
        self._scope = scope
        self._unparsed_expression: dict | None = None

    @property
    def underlying_node(self) -> Node:
        return self._node

    def add_unparsed_expression(self, expression: dict) -> None:
        assert self._unparsed_expression is None
        self._unparsed_expression = expression

    def analyze_expressions(self) -> None:
        if self._node.type == NodeType.VARIABLE and not self._node.expression:
            expression = self._node.variable_declaration.expression
            if expression:
                self._node.add_expression(expression)
        if self._unparsed_expression:
            expression = parse_yul(self._scope, self, self._unparsed_expression)
            if expression:
                self._node.add_expression(expression)

        if self._node.expression:
            if self._node.type == NodeType.VARIABLE:
                # Update the expression to be an assignement to the variable
                variable_declaration = self._node.variable_declaration
                if variable_declaration:
                    _expression = AssignmentOperation(
                        Identifier(self._node.variable_declaration),
                        self._node.expression,
                        AssignmentOperationType.ASSIGN,
                        variable_declaration.type,
                    )
                    _expression.set_offset(
                        self._node.expression.source_mapping, self._node.compilation_unit
                    )
                    self._node.add_expression(_expression, bypass_verif_empty=True)

            expression = self._node.expression
            read_var = ReadVar(expression)
            self._node.variables_read_as_expression = read_var.result()

            write_var = WriteVar(expression)
            self._node.variables_written_as_expression = write_var.result()

            find_call = FindCalls(expression)
            self._node.calls_as_expression = find_call.result()
            self._node.external_calls_as_expressions = [
                c for c in self._node.calls_as_expression if not isinstance(c.called, Identifier)
            ]
            self._node.internal_calls_as_expressions = [
                c for c in self._node.calls_as_expression if isinstance(c.called, Identifier)
            ]


def link_underlying_nodes(node1: YulNode, node2: YulNode) -> None:
    link_nodes(node1.underlying_node, node2.underlying_node)


def _name_to_yul_name(variable_name: str, yul_id: list[str]) -> str:
    """
    Translate the variable name to a unique yul name
    Within the same function, yul blocks can declare
    different variables with the same name
    We need to create unique name per variable
    to prevent collision during the SSA generation

    :param var:
    :param yul_id:
    :return:
    """
    return variable_name + f"_{'_'.join(yul_id)}"


class YulScope(metaclass=abc.ABCMeta):
    __slots__ = [
        "_contract",
        "_id",
        "_parent_func",
        "_yul_local_functions",
        "_yul_local_variables",
    ]

    def __init__(self, contract: Contract | None, yul_id: list[str], parent_func: Function) -> None:
        self._contract = contract
        self._id: list[str] = yul_id
        self._yul_local_variables: list[YulLocalVariable] = []
        self._yul_local_functions: list[YulFunction] = []
        self._parent_func: Function = parent_func

    @property
    def id(self) -> list[str]:
        return self._id

    @property
    def contract(self) -> Contract | None:
        return self._contract

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._parent_func.compilation_unit

    @property
    def parent_func(self) -> Function | None:
        return self._parent_func

    @property
    @abc.abstractmethod
    def function(self) -> Function:
        pass

    @abc.abstractmethod
    def new_node(self, node_type: NodeType, src: str | dict) -> YulNode:
        pass

    @property
    def file_scope(self) -> FileScope:
        return self._parent_func.file_scope

    def add_yul_local_variable(self, var: "YulLocalVariable") -> None:
        self._yul_local_variables.append(var)

    def get_yul_local_variable_from_name(self, variable_name: str) -> Optional["YulLocalVariable"]:
        return next(
            (
                v
                for v in self._yul_local_variables
                if v.underlying.name == _name_to_yul_name(variable_name, self.id)
            ),
            None,
        )

    def add_yul_local_function(self, func: "YulFunction") -> None:
        self._yul_local_functions.append(func)

    def get_yul_local_function_from_name(self, func_name: str) -> Optional["YulFunction"]:
        return next(
            (v for v in self._yul_local_functions if v.underlying.name == func_name),
            None,
        )


class YulLocalVariable:
    __slots__ = ["_root", "_variable"]

    def __init__(self, var: LocalVariable, root: YulScope, ast: dict) -> None:
        assert ast["nodeType"] == "YulTypedName"

        self._variable = var
        self._root = root

        # start initializing the underlying variable
        var.set_function(root.function)
        var.set_offset(ast["src"], root.compilation_unit)

        var.name = _name_to_yul_name(ast["name"], root.id)
        var.set_type(ElementaryType("uint256"))
        var.set_location("memory")

    @property
    def underlying(self) -> LocalVariable:
        return self._variable


class YulFunction(YulScope):
    __slots__ = ["_ast", "_entrypoint", "_function", "_nodes", "_root", "node_scope"]

    def __init__(
        self, func: Function, root: YulScope, ast: dict, node_scope: Function | Scope
    ) -> None:
        super().__init__(root.contract, root.id + [ast["name"]], parent_func=root.parent_func)

        assert ast["nodeType"] == "YulFunctionDefinition"

        self._function: Function = func
        self._root: YulScope = root
        self._ast: dict = ast

        # start initializing the underlying function

        func.name = ast["name"]
        func.set_visibility("private")
        if isinstance(func, SourceMapping):
            func.set_offset(ast["src"], root.compilation_unit)
        if isinstance(func, FunctionContract):
            func.set_contract(root.contract)
            func.set_contract_declarer(root.contract)
        func.compilation_unit = root.compilation_unit
        func.internal_scope = root.id
        func.is_implemented = True
        self.node_scope = node_scope

        self._nodes: list[YulNode] = []
        self._entrypoint = self.new_node(NodeType.ASSEMBLY, ast["src"])
        func.entry_point = self._entrypoint.underlying_node

        self.add_yul_local_function(self)

    @property
    def underlying(self) -> Function:
        return self._function

    @property
    def function(self) -> Function:
        return self._function

    @property
    def root(self) -> YulScope:
        return self._root

    def convert_body(self) -> None:
        node = self.new_node(NodeType.ENTRYPOINT, self._ast["src"])
        link_underlying_nodes(self._entrypoint, node)

        for param in self._ast.get("parameters", []):
            node = convert_yul(self, node, param, self.node_scope)
            self._function.add_parameters(
                self.get_yul_local_variable_from_name(param["name"]).underlying
            )

        for ret in self._ast.get("returnVariables", []):
            node = convert_yul(self, node, ret, self.node_scope)
            self._function.add_return(self.get_yul_local_variable_from_name(ret["name"]).underlying)

        convert_yul(self, node, self._ast["body"], self.node_scope)

    def parse_body(self) -> None:
        for node in self._nodes:
            node.analyze_expressions()
        for f in self._yul_local_functions:
            if f != self:
                f.parse_body()

    def new_node(self, node_type: NodeType, src: str) -> YulNode:
        if self._function:
            node = self._function.new_node(node_type, src, self.node_scope)
        else:
            raise SlitherException("standalone yul objects are not supported yet")

        yul_node = YulNode(node, self)
        self._nodes.append(yul_node)
        return yul_node


class YulBlock(YulScope):
    """
    A YulBlock represents a standalone yul component.
    For example an inline assembly block

    """

    __slots__ = ["_entrypoint", "_nodes", "_parent_func", "node_scope"]

    def __init__(
        self,
        contract: Contract | None,
        entrypoint: Node,
        yul_id: list[str],
        node_scope: Scope | Function,
    ) -> None:
        super().__init__(contract, yul_id, entrypoint.function)

        self._entrypoint: YulNode = YulNode(entrypoint, self)
        self._nodes: list[YulNode] = []
        self.node_scope = node_scope

    @property
    def entrypoint(self) -> YulNode:
        return self._entrypoint

    @property
    def function(self) -> Function:
        return self._parent_func

    @property
    def nodes(self) -> list[YulNode]:
        return self._nodes

    def new_node(self, node_type: NodeType, src: str | dict) -> YulNode:
        if self._parent_func:
            node = self._parent_func.new_node(node_type, src, self.node_scope)
        else:
            raise SlitherException("standalone yul objects are not supported yet")

        yul_node = YulNode(node, self)
        self._nodes.append(yul_node)
        return yul_node

    def convert(self, ast: dict) -> YulNode:
        yul_node = convert_yul(self, self._entrypoint, ast, self.node_scope)
        for f in self._yul_local_functions:
            f.parse_body()
        return yul_node

    def analyze_expressions(self) -> None:
        for node in self._nodes:
            node.analyze_expressions()


###################################################################################
###################################################################################
# region Block conversion
###################################################################################
###################################################################################


# The functions in this region, at a high level, will extract the control flow
# structures and metadata from the input AST. These include things like function
# definitions and local variables.
#
# Each function takes three parameters:
#     1)  root is the current YulScope, where you can find things like local variables
#     2)  parent is the previous YulNode, which you'll have to link to
#     3)  ast is a dictionary and is the current node in the Yul ast being converted
#
# Each function must return a single parameter:
#     1)  the new YulNode that the CFG ends at
#
# The entrypoint is the function at the end of this region, `convert_yul`, which
# dispatches to a specialized function based on a lookup dictionary.


def convert_yul_block(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    for statement in ast["statements"]:
        parent = convert_yul(root, parent, statement, node_scope)
    return parent


def convert_yul_function_definition(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    top_node_scope = node_scope
    while not isinstance(top_node_scope, Function):
        top_node_scope = top_node_scope.father

    func: FunctionTopLevel | FunctionContract
    if isinstance(top_node_scope, FunctionTopLevel):
        scope = root.file_scope
        func = FunctionTopLevel(root.compilation_unit, scope)
        # Note: we do not add the function in the scope
        # While its a top level function, it is not accessible outside of the function definition
        # In practice we should probably have a specific function type for function defined within a function
    else:
        func = FunctionContract(root.compilation_unit)

    func.function_language = FunctionLanguage.Yul
    yul_function = YulFunction(func, root, ast, node_scope)

    if root.contract:
        root.contract.add_function(func)

    root.compilation_unit.add_function(func)
    root.add_yul_local_function(yul_function)

    yul_function.convert_body()

    return parent


def convert_yul_variable_declaration(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    for variable_ast in ast["variables"]:
        parent = convert_yul(root, parent, variable_ast, node_scope)

    node = root.new_node(NodeType.EXPRESSION, ast["src"])
    node.add_unparsed_expression(ast)
    link_underlying_nodes(parent, node)

    return node


def convert_yul_assignment(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    node = root.new_node(NodeType.EXPRESSION, ast["src"])
    node.add_unparsed_expression(ast)
    link_underlying_nodes(parent, node)
    return node


def convert_yul_expression_statement(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    src = ast["src"]
    expression_ast = ast["expression"]

    expression = root.new_node(NodeType.EXPRESSION, src)
    expression.add_unparsed_expression(expression_ast)
    link_underlying_nodes(parent, expression)

    return expression


def convert_yul_if(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    # we're cheating and pretending that yul supports if/else so we can convert switch cleaner

    src = ast["src"]
    condition_ast = ast["condition"]
    true_body_ast = ast["body"]
    false_body_ast = ast.get("false_body")

    condition = root.new_node(NodeType.IF, src)
    end = root.new_node(NodeType.ENDIF, src)

    condition.add_unparsed_expression(condition_ast)

    true_body = convert_yul(root, condition, true_body_ast, node_scope)

    if false_body_ast:
        false_body = convert_yul(root, condition, false_body_ast, node_scope)
        link_underlying_nodes(false_body, end)
    else:
        link_underlying_nodes(condition, end)

    link_underlying_nodes(parent, condition)
    link_underlying_nodes(true_body, end)

    return end


def convert_yul_switch(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    """
    This is unfortunate. We don't really want a switch in our IR so we're going to
    translate it into a series of if/else statements.
    """
    cases_ast = ast["cases"]
    expression_ast = ast["expression"]

    # this variable stores the result of the expression so we don't accidentally compute it more than once
    switch_expr_var = f"switch_expr_{ast['src'].replace(':', '_')}"

    rewritten_switch = {
        "nodeType": "YulBlock",
        "src": ast["src"],
        "statements": [
            {
                "nodeType": "YulVariableDeclaration",
                "src": expression_ast["src"],
                "variables": [
                    {
                        "nodeType": "YulTypedName",
                        "src": expression_ast["src"],
                        "name": switch_expr_var,
                        "type": "",
                    },
                ],
                "value": expression_ast,
            },
        ],
    }

    last_if: dict | None = None

    default_ast = None

    for case_ast in cases_ast:
        body_ast = case_ast["body"]
        value_ast = case_ast["value"]

        if value_ast == "default":
            default_ast = case_ast
            continue

        current_if = {
            "nodeType": "YulIf",
            "src": case_ast["src"],
            "condition": {
                "nodeType": "YulFunctionCall",
                "src": case_ast["src"],
                "functionName": {
                    "nodeType": "YulIdentifier",
                    "src": case_ast["src"],
                    "name": "eq",
                },
                "arguments": [
                    {
                        "nodeType": "YulIdentifier",
                        "src": case_ast["src"],
                        "name": switch_expr_var,
                    },
                    value_ast,
                ],
            },
            "body": body_ast,
        }

        if last_if:
            last_if["false_body"] = current_if
        else:
            rewritten_switch["statements"].append(current_if)

        last_if = current_if

    if default_ast:
        body_ast = default_ast["body"]

        if last_if:
            last_if["false_body"] = body_ast
        else:
            rewritten_switch["statements"].append(body_ast)

    return convert_yul(root, parent, rewritten_switch, node_scope)


def convert_yul_for_loop(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    pre_ast = ast["pre"]
    condition_ast = ast["condition"]
    post_ast = ast["post"]
    body_ast = ast["body"]

    start_loop = root.new_node(NodeType.STARTLOOP, ast["src"])
    end_loop = root.new_node(NodeType.ENDLOOP, ast["src"])

    link_underlying_nodes(parent, start_loop)

    pre = convert_yul(root, start_loop, pre_ast, node_scope)

    condition = root.new_node(NodeType.IFLOOP, condition_ast["src"])
    condition.add_unparsed_expression(condition_ast)
    link_underlying_nodes(pre, condition)

    link_underlying_nodes(condition, end_loop)

    body = convert_yul(root, condition, body_ast, node_scope)

    post = convert_yul(root, body, post_ast, node_scope)

    link_underlying_nodes(post, condition)

    return end_loop


def convert_yul_break(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    break_ = root.new_node(NodeType.BREAK, ast["src"])
    link_underlying_nodes(parent, break_)
    return break_


def convert_yul_continue(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    continue_ = root.new_node(NodeType.CONTINUE, ast["src"])
    link_underlying_nodes(parent, continue_)
    return continue_


def convert_yul_leave(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    leave = root.new_node(NodeType.RETURN, ast["src"])
    link_underlying_nodes(parent, leave)
    return leave


def convert_yul_typed_name(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    local_var = LocalVariable()

    var = YulLocalVariable(local_var, root, ast)
    root.add_yul_local_variable(var)

    node = root.new_node(NodeType.VARIABLE, ast["src"])
    node.underlying_node.add_variable_declaration(local_var)
    link_underlying_nodes(parent, node)

    return node


def convert_yul_unsupported(
    root: YulScope, parent: YulNode, ast: dict, _node_scope: Function | Scope
) -> YulNode:
    raise SlitherException(
        f"no converter available for {ast['nodeType']} {json.dumps(ast, indent=2)}"
    )


def convert_yul(
    root: YulScope, parent: YulNode, ast: dict, node_scope: Function | Scope
) -> YulNode:
    return converters.get(ast["nodeType"], convert_yul_unsupported)(root, parent, ast, node_scope)


converters = {
    "YulBlock": convert_yul_block,
    "YulFunctionDefinition": convert_yul_function_definition,
    "YulVariableDeclaration": convert_yul_variable_declaration,
    "YulAssignment": convert_yul_assignment,
    "YulExpressionStatement": convert_yul_expression_statement,
    "YulIf": convert_yul_if,
    "YulSwitch": convert_yul_switch,
    "YulForLoop": convert_yul_for_loop,
    "YulBreak": convert_yul_break,
    "YulContinue": convert_yul_continue,
    "YulLeave": convert_yul_leave,
    "YulTypedName": convert_yul_typed_name,
}

# endregion
###################################################################################
###################################################################################

###################################################################################
###################################################################################
# region Expression parsing
###################################################################################
###################################################################################

"""
The functions in this region parse the AST into expressions.

Each function takes three parameters:
    1)  root is the same root as above
    2)  node is the CFG node which stores this expression
    3)  ast is the same ast as above

Each function must return a single parameter:
    1) The operation that was parsed, or None

The entrypoint is the function at the end of this region, `parse_yul`, which
dispatches to a specialized function based on a lookup dictionary.
"""


def _parse_yul_assignment_common(
    root: YulScope, node: YulNode, ast: dict, key: str
) -> Expression | None:
    lhs = [parse_yul(root, node, arg) for arg in ast[key]]
    rhs = parse_yul(root, node, ast["value"])

    return AssignmentOperation(
        vars_to_val(lhs), rhs, AssignmentOperationType.ASSIGN, vars_to_typestr(lhs)
    )


def parse_yul_variable_declaration(root: YulScope, node: YulNode, ast: dict) -> Expression | None:
    """
    We already created variables in the conversion phase, so just do
    the assignment
    """

    if "value" not in ast or not ast["value"]:
        return None

    return _parse_yul_assignment_common(root, node, ast, "variables")


def parse_yul_assignment(root: YulScope, node: YulNode, ast: dict) -> Expression | None:
    return _parse_yul_assignment_common(root, node, ast, "variableNames")


def parse_yul_function_call(root: YulScope, node: YulNode, ast: dict) -> Expression | None:
    args = [parse_yul(root, node, arg) for arg in ast["arguments"]]
    ident = parse_yul(root, node, ast["functionName"])

    if not isinstance(ident, Identifier):
        raise SlitherException("expected identifier from parsing function name")

    if isinstance(ident.value, YulBuiltin):
        name = ident.value.name
        if name in binary_ops:
            if name in ["shl", "shr", "sar"]:
                # lmao ok
                return BinaryOperation(args[1], args[0], binary_ops[name])

            return BinaryOperation(args[0], args[1], binary_ops[name])

        if name in unary_ops:
            return UnaryOperation(args[0], unary_ops[name])

        if name == "stop":
            name = "return"
            ident = Identifier(SolidityFunction(format_function_descriptor(name)))
            args = [
                Literal("0", ElementaryType("uint256")),
                Literal("0", ElementaryType("uint256")),
            ]

        else:
            ident = Identifier(SolidityFunction(format_function_descriptor(ident.value.name)))

    if isinstance(ident.value, Function):
        return CallExpression(ident, args, vars_to_typestr(ident.value.returns))
    if isinstance(ident.value, SolidityFunction):
        return CallExpression(ident, args, vars_to_typestr(ident.value.return_type))

    raise SlitherException(f"unexpected function call target type {type(ident.value)!s}")


def _check_for_state_variable_name(root: YulScope, potential_name: str) -> Identifier | None:
    root_function = root.function
    if isinstance(root_function, FunctionContract):
        var = root_function.contract_declarer.get_state_variable_from_name(potential_name)
        if var:
            return Identifier(var)
    return None


def _parse_yul_magic_suffixes(name: str, root: YulScope) -> Expression | None:
    # check for magic suffixes
    # TODO: the following leads to wrong IR
    # Currently SlithIR doesnt support raw access to memory
    # So things like .offset/.slot will return the variable
    # Instaed of the actual offset/slot
    if name.endswith(("_slot", ".slot")):
        potential_name = name[:-5]
        variable_found = _check_for_state_variable_name(root, potential_name)
        if variable_found:
            return variable_found
        var = root.function.get_local_variable_from_name(potential_name)
        if var and var.is_storage:
            return Identifier(var)
    if name.endswith(("_offset", ".offset")):
        potential_name = name[:-7]
        variable_found = _check_for_state_variable_name(root, potential_name)
        if variable_found:
            return variable_found
        var = root.function.get_local_variable_from_name(potential_name)
        if var and var.location == "calldata":
            return Identifier(var)
    if name.endswith(".length"):
        # TODO: length should create a new IP operation LENGTH var
        # The code below is an hotfix to allow slither to process length in yul
        # Until we have a better support
        potential_name = name[:-7]
        var = root.function.get_local_variable_from_name(potential_name)
        if var and var.location == "calldata":
            return Identifier(var)
    return None


def parse_yul_identifier(root: YulScope, _node: YulNode, ast: dict) -> Expression | None:
    name = ast["name"]

    if name in builtins:
        return Identifier(YulBuiltin(name))

    # check yul-scoped variable
    variable = root.get_yul_local_variable_from_name(name)
    if variable:
        return Identifier(variable.underlying)

    # check yul-scoped function
    func = root.get_yul_local_function_from_name(name)
    if func:
        return Identifier(func.underlying)

    # check yul-block scoped function
    if isinstance(root, YulFunction):
        yul_block = root.root

        # Iterate until we searched in all the scopes until the YulBlock scope
        while not isinstance(yul_block, YulBlock):
            func = yul_block.get_yul_local_function_from_name(name)
            if func:
                return Identifier(func.underlying)

            if isinstance(yul_block, YulFunction):
                yul_block = yul_block.root

        func = yul_block.get_yul_local_function_from_name(name)
        if func:
            return Identifier(func.underlying)

    # check function-scoped variables
    parent_func = root.parent_func
    if parent_func:
        local_variable = parent_func.get_local_variable_from_name(name)
        if local_variable:
            return Identifier(local_variable)

        if isinstance(parent_func, FunctionContract):
            # Variables must be looked from the contract declarer
            assert parent_func.contract_declarer
            state_variable = parent_func.contract_declarer.get_state_variable_from_name(name)
            if state_variable:
                return Identifier(state_variable)

    magic_suffix = _parse_yul_magic_suffixes(name, root)
    if magic_suffix:
        return magic_suffix

    ret, _ = find_top_level(name, root.file_scope)
    if ret:
        return Identifier(ret)

    raise SlitherException(f"unresolved reference to identifier {name}")


def parse_yul_literal(_root: YulScope, _node: YulNode, ast: dict) -> Expression | None:
    kind = ast["kind"]

    if kind == "string":
        # Solc 0.8.0 use value, 0.8.16 use hexValue - not sure when this changed was made
        if "value" in ast:
            value = ast["value"]
        else:
            value = ast["hexValue"]
    else:
        # number/bool
        value = ast["value"]

    if not kind:
        kind = "bool" if value in ["true", "false"] else "uint256"

    if kind == "number":
        kind = "uint256"

    return Literal(value, ElementaryType(kind))


def parse_yul_typed_name(root: YulScope, _node: YulNode, ast: dict) -> Expression | None:
    var = root.get_yul_local_variable_from_name(ast["name"])

    i = Identifier(var.underlying)
    i.type = var.underlying.type
    return i


def parse_yul_unsupported(_root: YulScope, _node: YulNode, ast: dict) -> Expression | None:
    raise SlitherException(f"no parser available for {ast['nodeType']} {json.dumps(ast, indent=2)}")


def parse_yul(root: YulScope, node: YulNode, ast: dict) -> Expression | None:
    op: Expression = parsers.get(ast["nodeType"], parse_yul_unsupported)(root, node, ast)
    if op:
        op.set_offset(ast["src"], root.compilation_unit)
    return op


parsers = {
    "YulVariableDeclaration": parse_yul_variable_declaration,
    "YulAssignment": parse_yul_assignment,
    "YulFunctionCall": parse_yul_function_call,
    "YulIdentifier": parse_yul_identifier,
    "YulTypedName": parse_yul_typed_name,
    "YulLiteral": parse_yul_literal,
}


# endregion
###################################################################################
###################################################################################


def vars_to_typestr(rets: list[Expression]) -> str:
    if len(rets) == 0:
        return ""
    if len(rets) == 1:
        return str(rets[0].type)
    return f"tuple({','.join(str(ret.type) for ret in rets)})"


def vars_to_val(
    vars_to_convert: list[Identifier],
) -> Identifier:
    if len(vars_to_convert) == 1:
        return vars_to_convert[0]
    return TupleExpression(vars_to_convert)



from pathlib import Path
from typing import TYPE_CHECKING
from slither.vyper_parsing.ast.types import (
    Module,
    FunctionDef,
    EventDef,
    EnumDef,
    StructDef,
    VariableDecl,
    ImportFrom,
    InterfaceDef,
    AnnAssign,
    Expr,
    Name,
    Arguments,
    Index,
    Subscript,
    Int,
    Arg,
)

from slither.vyper_parsing.declarations.event import EventVyper
from slither.vyper_parsing.declarations.struct import StructVyper
from slither.vyper_parsing.variables.state_variable import StateVariableVyper
from slither.vyper_parsing.declarations.function import FunctionVyper
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations import Contract, StructureContract, EnumContract, EventContract

from slither.core.variables.state_variable import StateVariable

if TYPE_CHECKING:
    from slither.vyper_parsing.vyper_compilation_unit import VyperCompilationUnit


class ContractVyper:
    def __init__(
        self, slither_parser: "VyperCompilationUnit", contract: Contract, module: Module
    ) -> None:
        self._contract: Contract = contract
        self._slither_parser: VyperCompilationUnit = slither_parser
        self._data = module
        # Vyper models only have one contract (aside from interfaces) and the name is the file path
        # We use the stem to make it a more user friendly name that is easy to query via canonical name
        self._contract.name = Path(module.name).stem
        self._contract.id = module.node_id
        self._is_analyzed: bool = False

        self._enumsNotParsed: list[EnumDef] = []
        self._structuresNotParsed: list[StructDef] = []
        self._variablesNotParsed: list[VariableDecl] = []
        self._eventsNotParsed: list[EventDef] = []
        self._functionsNotParsed: list[FunctionDef] = []

        self._structures_parser: list[StructVyper] = []
        self._variables_parser: list[StateVariableVyper] = []
        self._events_parser: list[EventVyper] = []
        self._functions_parser: list[FunctionVyper] = []

        self._parse_contract_items()

    @property
    def is_analyzed(self) -> bool:
        return self._is_analyzed

    def set_is_analyzed(self, is_analyzed: bool) -> None:
        self._is_analyzed = is_analyzed

    @property
    def underlying_contract(self) -> Contract:
        return self._contract

    def _parse_contract_items(self) -> None:
        for node in self._data.body:
            if isinstance(node, FunctionDef):
                self._functionsNotParsed.append(node)
            elif isinstance(node, EventDef):
                self._eventsNotParsed.append(node)
            elif isinstance(node, VariableDecl):
                self._variablesNotParsed.append(node)
            elif isinstance(node, EnumDef):
                self._enumsNotParsed.append(node)
            elif isinstance(node, StructDef):
                self._structuresNotParsed.append(node)
            elif isinstance(node, ImportFrom):
                # TOOD aliases
                # We create an `InterfaceDef` sense the compilation unit does not contain the actual interface
                # https://github.com/vyperlang/vyper/tree/master/vyper/builtins/interfaces
                if node.module == "vyper.interfaces":
                    interfaces = {
                        "ERC20Detailed": InterfaceDef(
                            src="-1:-1:-1",
                            node_id=-1,
                            name="ERC20Detailed",
                            body=[
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=-1,
                                    doc_string=None,
                                    name="name",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=-1,
                                        args=[],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Subscript(
                                        src="-1:-1:-1",
                                        node_id=-1,
                                        value=Name(src="-1:-1:-1", node_id=-1, id="String"),
                                        slice=Index(
                                            src="-1:-1:-1",
                                            node_id=-1,
                                            value=Int(src="-1:-1:-1", node_id=-1, value=1),
                                        ),
                                    ),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=-1,
                                            value=Name(src="-1:-1:-1", node_id=-1, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=-1,
                                    doc_string=None,
                                    name="symbol",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=-1,
                                        args=[],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Subscript(
                                        src="-1:-1:-1",
                                        node_id=-1,
                                        value=Name(src="-1:-1:-1", node_id=-1, id="String"),
                                        slice=Index(
                                            src="-1:-1:-1",
                                            node_id=-1,
                                            value=Int(src="-1:-1:-1", node_id=-1, value=1),
                                        ),
                                    ),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=-1,
                                            value=Name(src="-1:-1:-1", node_id=-1, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=-1,
                                    doc_string=None,
                                    name="decimals",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=-1,
                                        args=[],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=-1, id="uint8"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=-1,
                                            value=Name(src="-1:-1:-1", node_id=-1, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                            ],
                        ),
                        "ERC20": InterfaceDef(
                            src="-1:-1:-1",
                            node_id=1,
                            name="ERC20",
                            body=[
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=2,
                                    doc_string=None,
                                    name="totalSupply",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=3,
                                        args=[],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=7, id="uint256"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=4,
                                            value=Name(src="-1:-1:-1", node_id=5, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=9,
                                    doc_string=None,
                                    name="balanceOf",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=10,
                                        args=[
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=11,
                                                arg="_owner",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=12, id="address"
                                                ),
                                            )
                                        ],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=17, id="uint256"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=14,
                                            value=Name(src="-1:-1:-1", node_id=15, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=19,
                                    doc_string=None,
                                    name="allowance",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=20,
                                        args=[
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=21,
                                                arg="_owner",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=22, id="address"
                                                ),
                                            ),
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=24,
                                                arg="_spender",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=25, id="address"
                                                ),
                                            ),
                                        ],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=30, id="uint256"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=27,
                                            value=Name(src="-1:-1:-1", node_id=28, id="view"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=32,
                                    doc_string=None,
                                    name="transfer",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=33,
                                        args=[
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=34,
                                                arg="_to",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=35, id="address"
                                                ),
                                            ),
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=37,
                                                arg="_value",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=38, id="uint256"
                                                ),
                                            ),
                                        ],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=43, id="bool"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=40,
                                            value=Name(src="-1:-1:-1", node_id=41, id="nonpayable"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=45,
                                    doc_string=None,
                                    name="transferFrom",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=46,
                                        args=[
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=47,
                                                arg="_from",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=48, id="address"
                                                ),
                                            ),
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=50,
                                                arg="_to",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=51, id="address"
                                                ),
                                            ),
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=53,
                                                arg="_value",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=54, id="uint256"
                                                ),
                                            ),
                                        ],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=59, id="bool"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=56,
                                            value=Name(src="-1:-1:-1", node_id=57, id="nonpayable"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                                FunctionDef(
                                    src="-1:-1:-1",
                                    node_id=61,
                                    doc_string=None,
                                    name="approve",
                                    args=Arguments(
                                        src="-1:-1:-1",
                                        node_id=62,
                                        args=[
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=63,
                                                arg="_spender",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=64, id="address"
                                                ),
                                            ),
                                            Arg(
                                                src="-1:-1:-1",
                                                node_id=66,
                                                arg="_value",
                                                annotation=Name(
                                                    src="-1:-1:-1", node_id=67, id="uint256"
                                                ),
                                            ),
                                        ],
                                        default=None,
                                        defaults=[],
                                    ),
                                    returns=Name(src="-1:-1:-1", node_id=72, id="bool"),
                                    body=[
                                        Expr(
                                            src="-1:-1:-1",
                                            node_id=69,
                                            value=Name(src="-1:-1:-1", node_id=70, id="nonpayable"),
                                        )
                                    ],
                                    decorators=[],
                                    pos=None,
                                ),
                            ],
                        ),
                        "ERC165": [],
                        "ERC721": [],
                        "ERC4626": [],
                    }
                    self._data.body.append(interfaces[node.name])

            elif isinstance(node, InterfaceDef):
                # This needs to be done lazily as interfaces can refer to constant state variables
                contract = Contract(self._contract.compilation_unit, self._contract.file_scope)
                contract.set_offset(node.src, self._contract.compilation_unit)
                contract.is_interface = True

                contract_parser = ContractVyper(self._slither_parser, contract, node)
                self._contract.file_scope.contracts[contract.name] = contract

                self._slither_parser._underlying_contract_to_parser[contract] = contract_parser

            elif isinstance(node, AnnAssign):  # implements: ERC20
                pass  # TODO
            else:
                raise ValueError("Unknown contract node: ", node)

    def parse_enums(self) -> None:
        for enum in self._enumsNotParsed:
            name = enum.name
            canonicalName = self._contract.name + "." + enum.name
            values = [x.value.id for x in enum.body]
            new_enum = EnumContract(name, canonicalName, values)
            new_enum.set_contract(self._contract)
            new_enum.set_offset(enum.src, self._contract.compilation_unit)
            self._contract.enums_as_dict[name] = new_enum  # TODO solidity using canonicalName
        self._enumsNotParsed = []

    def parse_structs(self) -> None:
        for struct in self._structuresNotParsed:
            st = StructureContract(self._contract.compilation_unit)
            st.set_contract(self._contract)
            st.set_offset(struct.src, self._contract.compilation_unit)

            st_parser = StructVyper(st, struct)
            self._contract.structures_as_dict[st.name] = st
            self._structures_parser.append(st_parser)
            # Interfaces can refer to struct defs
            self._contract.file_scope.structures[st.name] = st

        self._structuresNotParsed = []

    def parse_state_variables(self) -> None:
        for varNotParsed in self._variablesNotParsed:
            var = StateVariable()
            var.set_contract(self._contract)
            var.set_offset(varNotParsed.src, self._contract.compilation_unit)

            var_parser = StateVariableVyper(var, varNotParsed)
            self._variables_parser.append(var_parser)

            assert var.name
            self._contract.variables_as_dict[var.name] = var
            self._contract.add_state_variables_ordered([var])
            # Interfaces can refer to constants
            self._contract.file_scope.variables[var.name] = var

        self._variablesNotParsed = []

    def parse_events(self) -> None:
        for event_to_parse in self._eventsNotParsed:
            event = EventContract()
            event.set_contract(self._contract)
            event.set_offset(event_to_parse.src, self._contract.compilation_unit)

            event_parser = EventVyper(event, event_to_parse)
            self._events_parser.append(event_parser)
            self._contract.events_as_dict[event.full_name] = event

    def parse_functions(self) -> None:
        for function in self._functionsNotParsed:
            func = FunctionContract(self._contract.compilation_unit)
            func.set_offset(function.src, self._contract.compilation_unit)
            func.set_contract(self._contract)
            func.set_contract_declarer(self._contract)

            func_parser = FunctionVyper(func, function, self)
            self._contract.add_function(func)
            self._contract.compilation_unit.add_function(func)
            self._functions_parser.append(func_parser)

        self._functionsNotParsed = []

    def analyze_state_variables(self):
        # Struct defs can refer to constant state variables
        for var_parser in self._variables_parser:
            var_parser.analyze(self._contract)

    def analyze(self) -> None:
        for struct_parser in self._structures_parser:
            struct_parser.analyze(self._contract)

        for event_parser in self._events_parser:
            event_parser.analyze(self._contract)

        for function in self._functions_parser:
            function.analyze_params()

        for function in self._functions_parser:
            function.analyze_content()

    def __hash__(self) -> int:
        return self._contract.id
